package jp.sourceforge.nicoro;

import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.http.Header;
import org.apache.http.HttpMessage;
import org.apache.http.HttpVersion;
import org.apache.http.conn.scheme.PlainSocketFactory;
import org.apache.http.conn.scheme.Scheme;
import org.apache.http.conn.scheme.SchemeRegistry;
import org.apache.http.conn.ssl.SSLSocketFactory;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.impl.conn.tsccm.ThreadSafeClientConnManager;
import org.apache.http.params.BasicHttpParams;
import org.apache.http.params.HttpParams;
import org.apache.http.params.HttpProtocolParams;
import org.apache.http.protocol.HTTP;
import org.json.JSONException;
import org.json.JSONObject;

import android.app.Activity;
import android.app.AlertDialog;
import android.content.DialogInterface;
import android.util.Log;
import android.webkit.CookieManager;

/**
 * 関数群
 */
public class Util {
	private static final boolean DEBUG_LOGD = Release.IS_DEBUG && true;
	private static final String LOG_TAG = "NicoRo";
	
	public static String getFirstHeaderValue(HttpMessage httpMessage, String name) {
		Header header = httpMessage.getFirstHeader(name);
		if (header != null) {
			return header.getValue();
		} else {
			return null;
		}
	}

	/**
	 * 指定パスのファイルからJSONを作成
	 * @param path
	 * @return
	 */
	public static JSONObject createJSONFromFile(String path) {
		FileReader reader = null;
		JSONObject json = null;
		try {
			reader = new FileReader(path);
			StringBuilder readString = new StringBuilder(1024);
			char[] buf = new char[1024];
			while (true) {
				int readLength = reader.read(buf, 0, buf.length);
				if (readLength < 0) {
					break;
				}
				readString.append(buf, 0, readLength);
			}
			json = new JSONObject(readString.toString());
		} catch (FileNotFoundException e) {
			Log.d(LOG_TAG, "", e);
		} catch (IOException e) {
			Log.d(LOG_TAG, "", e);
		} catch (JSONException e) {
			Log.d(LOG_TAG, "", e);
		} finally {
			if (reader != null) {
				try {
					reader.close();
				} catch (IOException e) {
					Log.d(LOG_TAG, "", e);
				}
			}
		}
		return json;
	}

	/**
	 * JSONを指定パスのファイルに保存
	 * @param path
	 * @param json
	 * @return trueで成功、falseで失敗
	 */
	public static boolean saveJSONToFile(String path, JSONObject json) {
		FileWriter writer = null;
		boolean result = false;
		try {
			writer = new FileWriter(path);
			writer.write(json.toString());
			result = true;
		} catch (IOException e) {
			Log.d(LOG_TAG, "", e);
		} finally {
			if (writer != null) {
				try {
					writer.close();
				} catch (IOException e) {
					Log.d(LOG_TAG, "", e);
				}
			}
		}
		return result;
	}
	
	public static void logHeaders(String tag, Header[] headers) {
		for (Header header : headers) {
			Log.d(tag, header.getName() + ": " + header.getValue());
		}
	}
	
	public static void showErrorDialog(final Activity activity,
			String errorMessage, final boolean tryFinishActivity) {
		AlertDialog.Builder alertDialogBuilder = new AlertDialog.Builder(activity);
		alertDialogBuilder.setTitle("エラー");
		alertDialogBuilder.setMessage(errorMessage);
		alertDialogBuilder.setCancelable(true);
		alertDialogBuilder.setNeutralButton("閉じる", new DialogInterface.OnClickListener() {
			@Override
			public void onClick(DialogInterface dialog, int which) {
				dialog.dismiss();
				if (tryFinishActivity) {
					activity.finish();
				}
			}
		});
		alertDialogBuilder.show();
	}

	public static String getCookieValueFromManager(String urlBase, String name) {
		CookieManager cookieManager = CookieManager.getInstance();
		String cookie = cookieManager.getCookie(urlBase);
		if (DEBUG_LOGD) {
			Log.d(LOG_TAG, "getCookieValueFromManager: urlBase=" + urlBase + " cookie=" + cookie);
		}
		Pattern pattern = Pattern.compile("(" + name + "=[^;\\n\\r]+)");
		Matcher matcher = pattern.matcher(cookie);
		if (matcher.find()) {
			return matcher.group(1);
		}
		return null;
	}

	public static DefaultHttpClient createHttpClient() {
		SchemeRegistry schemeRegistry = new SchemeRegistry();
		schemeRegistry.register(new Scheme("http", PlainSocketFactory.getSocketFactory(), 80));
		SSLSocketFactory sslSocketFactory = SSLSocketFactory.getSocketFactory();
		schemeRegistry.register(new Scheme("https", sslSocketFactory, 443));
	
		HttpParams httpParams = new BasicHttpParams();
		HttpProtocolParams.setVersion(httpParams, HttpVersion.HTTP_1_1);
		HttpProtocolParams.setContentCharset(httpParams, HTTP.UTF_8);
		DefaultHttpClient httpClient = new DefaultHttpClient(
				new ThreadSafeClientConnManager(httpParams, schemeRegistry),
				httpParams);
		return httpClient;
	}
}
